@@ -5,7 +5,6 @@ import (
 	"testing"
 
 	"github.com/cilium/ebpf/internal/testutils"
-	"github.com/cilium/ebpf/internal/unix"
 )
 
 func TestObjNameCharacters(t *testing.T) {
@@ -15,6 +14,7 @@ func TestObjNameCharacters(t *testing.T) {
 		"a-b":     false,
 		"yeah so": false,
 		"dot.":    objNameAllowsDot() == nil,
+		"Capital": true,
 	} {
 		result := strings.IndexFunc(in, invalidBPFObjNameChar) == -1
 		if result != valid {
@@ -23,14 +23,8 @@ func TestObjNameCharacters(t *testing.T) {
 	}
 }
 
-func TestObjName(t *testing.T) {
-	name := newBPFObjName("more_than_16_characters_long")
-	if name[len(name)-1] != 0 {
-		t.Error("newBPFObjName doesn't null terminate")
-	}
-	if len(name) != unix.BPF_OBJ_NAME_LEN {
-		t.Errorf("Name is %d instead of %d bytes long", len(name), unix.BPF_OBJ_NAME_LEN)
-	}
+func TestHaveBatchAPI(t *testing.T) {
+	testutils.CheckFeatureTest(t, haveBatchAPI)
 }
 
 func TestHaveObjName(t *testing.T) {
@@ -48,3 +42,19 @@ func TestHaveNestedMaps(t *testing.T) {
 func TestHaveMapMutabilityModifiers(t *testing.T) {
 	testutils.CheckFeatureTest(t, haveMapMutabilityModifiers)
 }
+
+func TestHaveMmapableMaps(t *testing.T) {
+	testutils.CheckFeatureTest(t, haveMmapableMaps)
+}
+
+func TestHaveInnerMaps(t *testing.T) {
+	testutils.CheckFeatureTest(t, haveInnerMaps)
+}
+
+func TestHaveProbeReadKernel(t *testing.T) {
+	testutils.CheckFeatureTest(t, haveProbeReadKernel)
+}
+
+func TestHaveBPFToBPFCalls(t *testing.T) {
+	testutils.CheckFeatureTest(t, haveBPFToBPFCalls)
+}
