@@ -13,21 +13,22 @@ struct {
 	__uint(type, BPF_MAP_TYPE_PROG_ARRAY);
 	__uint(key_size, sizeof(uint32_t));
 	__uint(max_entries, 2);
-	__array(values, int ());
+	__array(values, int());
 } prog_array_init __section(".maps") = {
-	.values = {
-    // Skip index 0 to exercise empty array slots.
-		[1] = &tail_1,
-	},
+	.values =
+		{
+			// Skip index 0 to exercise empty array slots.
+			[1] = &tail_1,
+		},
 };
 
 int __section("socket/main") tail_main(void *ctx) {
-  // If prog_array_init is correctly populated, the tail call
-  // will succeed and the program will continue in tail_1 and
-  // not return here.
-  tail_call(ctx, &prog_array_init, 1);
+	// If prog_array_init is correctly populated, the tail call
+	// will succeed and the program will continue in tail_1 and
+	// not return here.
+	tail_call(ctx, &prog_array_init, 1);
 
-  return 0;
+	return 0;
 }
 
 // Inner map with a single possible entry.
@@ -46,10 +47,11 @@ struct {
 	__uint(value_size, sizeof(uint32_t));
 	__array(values, typeof(inner_map));
 } outer_map_init __section(".maps") = {
-	.values = {
-		// Skip index 0 to exercise empty array slots.
-		[1] = &inner_map,
-	},
+	.values =
+		{
+			// Skip index 0 to exercise empty array slots.
+			[1] = &inner_map,
+		},
 };
 
 #else
