@@ -29,6 +29,8 @@ struct bpf_map_def {
 
 static void *(*map_lookup_elem)(const void *map, const void *key) = (void *)1;
 
+static long (*map_update_elem)(const void *map, const void *key, const void *value, uint64_t flags) = (void *)2;
+
 static long (*trace_printk)(const char *fmt, uint32_t fmt_size, ...) = (void *)6;
 
 static long (*tail_call)(void *ctx, void *prog_array_map, uint32_t index) = (void *)12;
@@ -36,3 +38,5 @@ static long (*tail_call)(void *ctx, void *prog_array_map, uint32_t index) = (voi
 static int (*perf_event_output)(const void *ctx, const void *map, uint64_t index, const void *data, uint64_t size) = (void *)25;
 
 static uint32_t (*get_smp_processor_id)(void) = (void *)8;
+
+static long (*for_each_map_elem)(const void *map, void *callback_fn, void *callback_ctx, uint64_t flags) = (void *)164;
